#include "util.h"
#include "client.h"
#include "wm.h"
#include <stdarg.h>
#include <stddef.h>
#include <stdint.h>
#include <setjmp.h>
#include <cmocka.h>

// store a copy of the nodes created in test2_setup, to check if any
// of the tests modified them
static NodeArray *test_group2_nodes_copy = NULL;

static void test_wm_nodearray_push(void **state)
{
    TreeNode *node = wm_treenode_new(NODE_VERTICAL, NULL);
    TreeNode *node1 = wm_treenode_new(NODE_VERTICAL, node);

    assert_non_null(node);
    assert_non_null(node1);

    node1->id = 123;

    assert_null(node->parent);
    assert_ptr_equal(node1->parent, node);

    wm_nodearray_push(node->children, node1);

    assert_int_equal(node->children->size, 1);
    assert_ptr_equal(node->children->nodes[0], node1);
    assert_int_equal(node1->id, 123);
    assert_int_equal(node->children->nodes[0]->id, 123);

    wm_treenode_free(node);
    wm_treenode_free(node1);
}

static void test_wm_treenode_new(void **state)
{
    TreeNode *node = wm_treenode_new(NODE_VERTICAL, NULL);

    assert_non_null(node);
    assert_null(node->client);
    assert_non_null(node->children->nodes);
    assert_int_equal(node->children->size, 0);
    assert_int_equal(node->type, NODE_VERTICAL);

    wm_treenode_free(node);
}

static void test_wm_treenode_split_space(void **state)
{
    /* vertical */
    TreeNode *parent = wm_treenode_new(NODE_VERTICAL, NULL);
    assert_non_null(parent);

    parent->pos.x = 0;
    parent->pos.y = 0;
    parent->pos.w = 200;
    parent->pos.h = 200;

    TreeNode *child1 = wm_treenode_new(NODE_CLIENT, parent);
    TreeNode *child2 = wm_treenode_new(NODE_CLIENT, parent);

    assert_non_null(child1);
    assert_non_null(child2);

    wm_treenode_split_space(parent, &child1->pos, &child2->pos);

    assert_int_equal(child1->pos.x, 0);
    assert_int_equal(child1->pos.y, 0);

    assert_int_equal(child1->pos.w, 100);
    assert_int_equal(child1->pos.h, 200);

    assert_int_equal(child2->pos.x, 100);
    assert_int_equal(child2->pos.y, 0);

    assert_int_equal(child2->pos.w, 100);
    assert_int_equal(child2->pos.h, 200);

    wm_treenode_free(parent);
    wm_treenode_free(child1);
    wm_treenode_free(child2);

    /* horizontal */
    parent = wm_treenode_new(NODE_HORIZONTAL, NULL);
    assert_non_null(parent);

    parent->pos.x = 0;
    parent->pos.y = 0;
    parent->pos.w = 200;
    parent->pos.h = 200;

    child1 = wm_treenode_new(NODE_CLIENT, parent);
    child2 = wm_treenode_new(NODE_CLIENT, parent);

    assert_non_null(child1);
    assert_non_null(child2);

    wm_treenode_split_space(parent, &child1->pos, &child2->pos);

    assert_int_equal(child1->pos.x, 0);
    assert_int_equal(child1->pos.y, 0);

    assert_int_equal(child1->pos.w, 200);
    assert_int_equal(child1->pos.h, 100);

    assert_int_equal(child2->pos.x, 0);
    assert_int_equal(child2->pos.y, 100);

    assert_int_equal(child2->pos.w, 200);
    assert_int_equal(child2->pos.h, 100);

    wm_treenode_free(parent);
    wm_treenode_free(child1);
    wm_treenode_free(child2);
}

static void test_wm_all_nodes_to_array(void **state)
{
    TreeNode *root = wm_treenode_new(NODE_VERTICAL, NULL);
    TreeNode *node1 = wm_treenode_new(NODE_CLIENT, NULL);
    TreeNode *node2 = wm_treenode_new(NODE_CLIENT, NULL);
    TreeNode *node3 = wm_treenode_new(NODE_CLIENT, NULL);

    assert_non_null(root);
    assert_non_null(node1);
    assert_non_null(node2);
    assert_non_null(node3);

    const unsigned long ids[] = {1, 2, 3, 4};

    root->id = ids[0];
    node1->id = ids[1];
    node2->id = ids[2];
    node3->id = ids[3];

    wm_nodearray_push(root->children, node1);
    wm_nodearray_push(root->children, node2);
    wm_nodearray_push(root->children, node3);

    assert_int_equal(root->children->size, 3);
    assert_ptr_equal(root->children->nodes[0], node1);
    assert_ptr_equal(root->children->nodes[1], node2);
    assert_ptr_equal(root->children->nodes[2], node3);

    NodeArray *all_nodes = wm_all_nodes_to_array(root);
    assert_non_null(all_nodes);

    assert_int_equal(all_nodes->size, 4);
    assert_in_set(((TreeNode*)all_nodes->nodes[0])->id, ids, 4);
    assert_in_set(((TreeNode*)all_nodes->nodes[1])->id, ids, 4);
    assert_in_set(((TreeNode*)all_nodes->nodes[2])->id, ids, 4);
    assert_in_set(((TreeNode*)all_nodes->nodes[3])->id, ids, 4);

    wm_treenode_free(root);
    wm_treenode_free(node1);
    wm_treenode_free(node2);
    wm_treenode_free(node3);
    wm_nodearray_free(all_nodes);
}

static void test_wm_find_client_nodes(void **state)
{
    TreeNode *root = wm_treenode_new(NODE_VERTICAL, NULL);
    TreeNode *node1 = wm_treenode_new(NODE_CLIENT, NULL);
    TreeNode *node2 = wm_treenode_new(NODE_CLIENT, NULL);

    assert_non_null(root);
    assert_non_null(node1);
    assert_non_null(node2);

    node1->client = (Client*)0x123;
    node2->client = (Client*)0x456;

    const unsigned long ids[] = {123, 456};

    node1->id = ids[0];
    node2->id = ids[1];

    wm_nodearray_push(root->children, node1);
    wm_nodearray_push(root->children, node2);

    assert_int_equal(root->children->size, 2);
    assert_ptr_equal(root->children->nodes[0], node1);
    assert_ptr_equal(root->children->nodes[1], node2);

    NodeArray *client_nodes = wm_treenode_find_client_nodes(root);
    assert_non_null(client_nodes);

    assert_int_equal(client_nodes->size, 2);
    assert_in_set((client_nodes->nodes[0])->id, ids, 2);
    assert_in_set((client_nodes->nodes[1])->id, ids, 2);

    wm_treenode_free(root);
    wm_treenode_free(node1);
    wm_treenode_free(node2);
    wm_nodearray_free(client_nodes);
}

static void test_wm_treenode_remove_client1(void **state)
{
    TreeNode *root = wm_treenode_new(NODE_VERTICAL, NULL);
    TreeNode *parent_node1 = wm_treenode_new(NODE_VERTICAL, root);
    TreeNode *parent_node2 = wm_treenode_new(NODE_HORIZONTAL, root);
    TreeNode *client_node1 = wm_treenode_new(NODE_CLIENT, parent_node1);
    TreeNode *client_node2 = wm_treenode_new(NODE_CLIENT, parent_node1);
    TreeNode *client_node3 = wm_treenode_new(NODE_CLIENT, parent_node2);
    TreeNode *client_node4 = wm_treenode_new(NODE_CLIENT, parent_node2);

    assert_non_null(root);
    assert_non_null(parent_node1);
    assert_non_null(parent_node2);
    assert_non_null(client_node1);
    assert_non_null(client_node2);
    assert_non_null(client_node3);
    assert_non_null(client_node4);

    wm_nodearray_push(root->children, parent_node1);
    wm_nodearray_push(root->children, parent_node2);

    wm_nodearray_push(parent_node1->children, client_node1);
    wm_nodearray_push(parent_node1->children, client_node2);

    wm_nodearray_push(parent_node2->children, client_node3);
    wm_nodearray_push(parent_node2->children, client_node4);

    Workspace ws = (Workspace) {
        .tree = root,
        .monitor = &(Monitor) {0},
    };
    Client client1 = (Client) {.ws = &ws, .name = ""};
    Client client2 = (Client) {.ws = &ws, .name = ""};
    Client client3 = (Client) {.ws = &ws, .name = ""};
    Client client4 = (Client) {.ws = &ws, .name = ""};

    client_node1->client = &client1;
    client_node2->client = &client2;
    client_node3->client = &client3;
    client_node4->client = &client4;

    root->pos.x = 0;
    root->pos.y = 0;
    root->pos.w = 1000;
    root->pos.w = 1000;

    wm_treenode_split_space(root, &parent_node1->pos, &parent_node2->pos);
    wm_treenode_split_space(parent_node1, &client_node1->pos, &client_node2->pos);
    wm_treenode_split_space(parent_node2, &client_node3->pos, &client_node4->pos);

    TreeNode *new_client_node = wm_treenode_remove_client(&(Wm) {0}, root, &client4);
    assert_non_null(new_client_node);

    assert_ptr_equal(new_client_node, parent_node2);
    assert_int_equal(parent_node2->type, NODE_CLIENT);
    assert_ptr_equal(parent_node2->client, &client3);
    assert_int_equal(parent_node2->children->size, 0);

    wm_treenode_free(root);
    wm_treenode_free(parent_node1);
    wm_treenode_free(parent_node2);
    wm_treenode_free(client_node1);
    wm_treenode_free(client_node2);
}

static void test_wm_treenode_remove_client2(void **state)
{
    TreeNode *root = wm_treenode_new(NODE_VERTICAL, NULL);
    TreeNode *parent_node1 = wm_treenode_new(NODE_VERTICAL, root);
    TreeNode *client_node1 = wm_treenode_new(NODE_CLIENT, root);
    TreeNode *client_node2 = wm_treenode_new(NODE_CLIENT, parent_node1);
    TreeNode *client_node3 = wm_treenode_new(NODE_CLIENT, parent_node1);

    assert_non_null(root);
    assert_non_null(parent_node1);
    assert_non_null(client_node1);
    assert_non_null(client_node2);
    assert_non_null(client_node3);

    TreeNode parent_node1_val = *parent_node1;

    wm_nodearray_push(root->children, parent_node1);
    wm_nodearray_push(root->children, client_node1);

    wm_nodearray_push(parent_node1->children, client_node2);
    wm_nodearray_push(parent_node1->children, client_node3);

    Workspace ws = (Workspace) {
        .tree = root,
        .monitor = &(Monitor) {0},
    };
    Client client1 = (Client) {.ws = &ws, .name = ""};
    Client client2 = (Client) {.ws = &ws, .name = ""};
    Client client3 = (Client) {.ws = &ws, .name = ""};

    client_node1->client = &client1;
    client_node2->client = &client2;
    client_node3->client = &client3;

    root->pos.x = 0;
    root->pos.y = 0;
    root->pos.w = 1000;
    root->pos.w = 1000;

    wm_treenode_split_space(root, &parent_node1->pos, &client_node1->pos);
    wm_treenode_split_space(parent_node1, &client_node2->pos, &client_node3->pos);

    TreeNode *new_client_node = wm_treenode_remove_client(&(Wm) {0}, root, &client1);
    assert_non_null(new_client_node);

    assert_ptr_equal(new_client_node, client_node2);
    assert_int_equal(root->id, parent_node1_val.id);
    assert_ptr_equal(root->children, parent_node1_val.children);
    assert_int_equal(root->children->size, 2);
    assert_ptr_equal(root->children->nodes[0], client_node2);
    assert_ptr_equal(root->children->nodes[1], client_node3);
    assert_ptr_equal(client_node2->parent, root);
    assert_ptr_equal(client_node3->parent, root);

    wm_treenode_free(root);
    wm_treenode_free(client_node2);
    wm_treenode_free(client_node3);
}

static void test_wm_postorder_traversal(void **state)
{
    TreeNode *node1 = (TreeNode*)*state;

    NodeArray *postorder = wm_postorder_traversal(node1);

    assert_non_null(postorder);
    assert_int_equal(postorder->size, 8);

    assert_int_equal(postorder->nodes[0]->id, 4);
    assert_int_equal(postorder->nodes[1]->id, 5);
    assert_int_equal(postorder->nodes[2]->id, 6);
    assert_int_equal(postorder->nodes[3]->id, 2);
    assert_int_equal(postorder->nodes[4]->id, 7);
    assert_int_equal(postorder->nodes[5]->id, 8);
    assert_int_equal(postorder->nodes[6]->id, 3);
    assert_int_equal(postorder->nodes[7]->id, 1);

    wm_nodearray_free(postorder);
}

static void test_wm_treenode_lmd(void **state)
{
    TreeNode *node1 = (TreeNode*)*state;
    TreeNode *node2 = node1->children->nodes[0];
    TreeNode *node3 = node1->children->nodes[1];
    TreeNode *node4 = node2->children->nodes[0];
    TreeNode *node5 = node2->children->nodes[1];
    TreeNode *node6 = node2->children->nodes[2];
    TreeNode *node7 = node3->children->nodes[0];
    TreeNode *node8 = node3->children->nodes[1];

    TreeNode *lmd1 = wm_treenode_lmd(node1);
    TreeNode *lmd2 = wm_treenode_lmd(node2);
    TreeNode *lmd3 = wm_treenode_lmd(node3);
    TreeNode *lmd4 = wm_treenode_lmd(node4);
    TreeNode *lmd5 = wm_treenode_lmd(node5);
    TreeNode *lmd6 = wm_treenode_lmd(node6);
    TreeNode *lmd7 = wm_treenode_lmd(node7);
    TreeNode *lmd8 = wm_treenode_lmd(node8);

    assert_non_null(lmd1);
    assert_non_null(lmd2);
    assert_non_null(lmd3);
    assert_non_null(lmd4);
    assert_non_null(lmd5);
    assert_non_null(lmd6);
    assert_non_null(lmd7);
    assert_non_null(lmd8);

    assert_ptr_equal(lmd1, node4);
    assert_ptr_equal(lmd2, node4);
    assert_ptr_equal(lmd3, node7);
    assert_ptr_equal(lmd4, node4);
    assert_ptr_equal(lmd5, node5);
    assert_ptr_equal(lmd6, node6);
    assert_ptr_equal(lmd7, node7);
    assert_ptr_equal(lmd8, node8);
}

static void test_wm_is_treenode_keyroot(void **state)
{
    TreeNode *node1 = (TreeNode*)*state;

    NodeArray *postorder = wm_postorder_traversal(node1);

    assert_non_null(postorder);
    assert_int_equal(postorder->size, 8);

    assert_int_equal(wm_is_treenode_keyroot(postorder->nodes[0]), false);
    assert_int_equal(wm_is_treenode_keyroot(postorder->nodes[1]), true);
    assert_int_equal(wm_is_treenode_keyroot(postorder->nodes[2]), true);
    assert_int_equal(wm_is_treenode_keyroot(postorder->nodes[3]), false);
    assert_int_equal(wm_is_treenode_keyroot(postorder->nodes[4]), false);
    assert_int_equal(wm_is_treenode_keyroot(postorder->nodes[5]), true);
    assert_int_equal(wm_is_treenode_keyroot(postorder->nodes[6]), true);
    assert_int_equal(wm_is_treenode_keyroot(postorder->nodes[7]), true);

    wm_nodearray_free(postorder);
}

static void test_wm_treenode_all_lmds(void **state)
{
    TreeNode *node1 = (TreeNode*)*state;

    NodeArray *lmds = wm_treenode_all_lmds(node1);

    assert_non_null(lmds);
    assert_int_equal(lmds->size, 8);

    assert_int_equal(lmds->nodes[0]->id, 4);
    assert_int_equal(lmds->nodes[1]->id, 5);
    assert_int_equal(lmds->nodes[2]->id, 6);
    assert_int_equal(lmds->nodes[3]->id, 4);
    assert_int_equal(lmds->nodes[4]->id, 7);
    assert_int_equal(lmds->nodes[5]->id, 8);
    assert_int_equal(lmds->nodes[6]->id, 7);
    assert_int_equal(lmds->nodes[7]->id, 4);

    wm_nodearray_free(lmds);
}

static void test_wm_treenode_all_keyroots(void **state)
{
    TreeNode *node1 = (TreeNode*)*state;

    NodeArray *keyroots = wm_treenode_all_keyroots(node1);

    assert_non_null(keyroots);
    assert_int_equal(keyroots->size, 5);

    assert_int_equal(keyroots->nodes[0]->id, 5);
    assert_int_equal(keyroots->nodes[1]->id, 6);
    assert_int_equal(keyroots->nodes[2]->id, 8);
    assert_int_equal(keyroots->nodes[3]->id, 3);
    assert_int_equal(keyroots->nodes[4]->id, 1);

    wm_nodearray_free(keyroots);
}

static int test2_setup(void **state)
{
    TreeNode *node1 = wm_treenode_new(NODE_VERTICAL, NULL);
    TreeNode *node2 = wm_treenode_new(NODE_VERTICAL, node1);
    TreeNode *node3 = wm_treenode_new(NODE_VERTICAL, node1);
    TreeNode *node4 = wm_treenode_new(NODE_VERTICAL, node2);
    TreeNode *node5 = wm_treenode_new(NODE_VERTICAL, node2);
    TreeNode *node6 = wm_treenode_new(NODE_VERTICAL, node2);
    TreeNode *node7 = wm_treenode_new(NODE_VERTICAL, node3);
    TreeNode *node8 = wm_treenode_new(NODE_VERTICAL, node3);

    assert_non_null(node1);
    assert_non_null(node2);
    assert_non_null(node3);
    assert_non_null(node4);
    assert_non_null(node5);
    assert_non_null(node6);
    assert_non_null(node7);
    assert_non_null(node8);

    wm_nodearray_push(node1->children, node2);
    wm_nodearray_push(node1->children, node3);
    wm_nodearray_push(node2->children, node4);
    wm_nodearray_push(node2->children, node5);
    wm_nodearray_push(node2->children, node6);
    wm_nodearray_push(node3->children, node7);
    wm_nodearray_push(node3->children, node8);

    node1->id = 1;
    node2->id = 2;
    node3->id = 3;
    node4->id = 4;
    node5->id = 5;
    node6->id = 6;
    node7->id = 7;
    node8->id = 8;

    assert_null(node1->parent);
    assert_ptr_equal(node2->parent, node1);
    assert_ptr_equal(node3->parent, node1);
    assert_ptr_equal(node4->parent, node2);
    assert_ptr_equal(node5->parent, node2);
    assert_ptr_equal(node6->parent, node2);
    assert_ptr_equal(node7->parent, node3);
    assert_ptr_equal(node8->parent, node3);

    assert_int_equal(node1->children->size, 2);
    assert_int_equal(node2->children->size, 3);
    assert_int_equal(node3->children->size, 2);

    assert_int_equal(node1->id, 1);
    assert_int_equal(node1->children->nodes[0]->id, 2);
    assert_int_equal(node1->children->nodes[1]->id, 3);
    assert_int_equal(node2->children->nodes[0]->id, 4);
    assert_int_equal(node2->children->nodes[1]->id, 5);
    assert_int_equal(node2->children->nodes[2]->id, 6);
    assert_int_equal(node3->children->nodes[0]->id, 7);
    assert_int_equal(node3->children->nodes[1]->id, 8);

    *state = node1;

    NodeArray *nodes = wm_all_nodes_to_array((TreeNode*)*state);
    assert_non_null(nodes);

    test_group2_nodes_copy = wm_nodearray_new();
    for (size_t i = 0; i < nodes->size; i++) {
        TreeNode *node_copy = malloc(sizeof(TreeNode));
        assert_non_null(node_copy);
        *node_copy = *(nodes->nodes)[i];
        wm_nodearray_push(test_group2_nodes_copy, node_copy);
    }

    wm_nodearray_free(nodes);

    return 0;
}

static int test2_teardown(void **state)
{
    NodeArray *nodes = wm_all_nodes_to_array((TreeNode*)*state);

    assert_non_null(test_group2_nodes_copy);

    // check if any of the original nodes were modified
    for (size_t i = 0; i < test_group2_nodes_copy->size; i++) {
        TreeNode *node_copy = ((TreeNode**)test_group2_nodes_copy->nodes)[i];
        assert_memory_equal(node_copy, nodes->nodes[i], sizeof(TreeNode));
    }

    for (size_t i = 0; i < nodes->size; i++) {
        wm_treenode_free(nodes->nodes[i]);
        free((TreeNode*)test_group2_nodes_copy->nodes[i]);
    }

    wm_nodearray_free(nodes);
    wm_nodearray_free(test_group2_nodes_copy);

    return 0;
}

int main(void)
{
    const struct CMUnitTest test_group1[] = {
        cmocka_unit_test(test_wm_nodearray_push),
        cmocka_unit_test(test_wm_treenode_new),
        cmocka_unit_test(test_wm_treenode_split_space),
        cmocka_unit_test(test_wm_all_nodes_to_array),
        cmocka_unit_test(test_wm_find_client_nodes),
        cmocka_unit_test(test_wm_treenode_remove_client1),
        cmocka_unit_test(test_wm_treenode_remove_client2),
    };

    const struct CMUnitTest test_group2[] = {
        cmocka_unit_test(test_wm_postorder_traversal),
        cmocka_unit_test(test_wm_treenode_lmd),
        cmocka_unit_test(test_wm_is_treenode_keyroot),
        cmocka_unit_test(test_wm_treenode_all_lmds),
        cmocka_unit_test(test_wm_treenode_all_keyroots),
    };

    cmocka_run_group_tests(test_group1, NULL, NULL);
    cmocka_run_group_tests(test_group2, test2_setup, test2_teardown);
}
